""" Code to test the loading of OSIRIS Level 1 data from the standard HDF4 / onyx files.
"""
from unittest import TestCase
from typing import DefaultDict, List
from collections import defaultdict
import os
import numpy as np
from osirisl1services.readlevel1 import Orbit
from osirisl1services.readlevel1 import (open_level1_attitude,
                                         open_level1_housekeeping,
                                         open_level1_spectrograph,
                                         open_level1_ir,
                                         open_level1b_ir,
                                         open_level1_ecmwf)
import osl1 as l1
from osirisl1services.spatialrotation import ECI, Geodetic
from osirisl1services.readlevel1 import MIN_RADIATION_HIT
from osirisl1services.pyhdfadapter import HDF4VS


class TestReadLevel1(TestCase):
    def setUp(self):
        """ List of scans to be used for testing. TODO: Expand this list
        """
        self.scannos = [6432012,
                        72176000,  # catches #9
                        5579049,  # catches #27
                        5487016]  # catches #27

    def test_caching(self):
        """ Test that caching ensures that the second call to the same orbit (with a different scan) that the first uses
            the cache.
        """
        scanno = 6432012
        open_level1_spectrograph(scanno=scanno)
        hits = HDF4VS.cache_info().hits
        # First ensure that the next scan is not in the scan list
        self.assertNotIn(scanno + 1, self.scannos)
        open_level1_spectrograph(scanno=scanno + 1)
        self.assertGreater(HDF4VS.cache_info().hits, hits)

    def test_level1_attitude_does_not_contain_scan(self):
        """ Test scan not in attitude file raises KeyError. """

        scanno = 6432099
        with self.assertRaises(KeyError):
            open_level1_spectrograph(scanno=scanno)

    def test_level1_attitude_info(self):
        """ Test attitude scan info data compared to the scan info returned by c++ level1 services.
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno)
            scan_info_swig = l1.osl1.GetScanInfo(scanno)[1]
            self.assertEqual(attitude.StartMJD, scan_info_swig.StartMJD, msg=f'ScanNo: {scanno}')
            self.assertEqual(attitude.EndMJD, scan_info_swig.EndMJD, msg=f'ScanNo: {scanno}')
            self.assertEqual(attitude.StartMJD, scan_info_swig.StartMJD, msg=f'ScanNo: {scanno}')
            self.assertEqual(attitude.ACSState, scan_info_swig.ACSState, msg=f'ScanNo: {scanno}')
            # TODO: reinstate once ARGPackages/OSIRISLevel1Services#2 is fixed
            # self.assertEqual(attitude.Direction, scan_info_swig.Direction)
            self.assertEqual(attitude.MinAltitude, scan_info_swig.MinAltitude, msg=f'ScanNo: {scanno}')
            self.assertEqual(attitude.MaxAltitude, scan_info_swig.MaxAltitude, msg=f'ScanNo: {scanno}')

    def test_level1_attitude_position(self):
        """ Test attitude position compared to the position returned by c++ level1 services.
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno)
            for mjd, att in attitude.groupby('mjdorbitstate'):
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                self.assertAlmostEqual(position_eci.x, float(att.sel(xyz='x').position), places=5,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(position_eci.y, float(att.sel(xyz='y').position), places=5,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(position_eci.z, float(att.sel(xyz='z').position), places=5,
                                       msg=f'ScanNo: {scanno}')

    def test_level1_attitude_velocity(self):
        """ Test attitude velocity compared to the velocity returned by c++ level1 services.
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno)
            for mjd, att in attitude.groupby('mjdorbitstate'):
                velocity_eci = ECI(*l1.osl1.GetOdinVelocity(mjd)[1])
                self.assertAlmostEqual(velocity_eci.x, float(att.sel(xyz='x').velocity), msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(velocity_eci.y, float(att.sel(xyz='y').velocity), msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(velocity_eci.z, float(att.sel(xyz='z').velocity), msg=f'ScanNo: {scanno}')

    def test_level1_attitude_tangentpoint(self):
        """ Test attitude SSC lat/lon/height compared to the lat/lon/height returned by c++ level1 services.
            These numbers are only within two degrees!
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_ssc_los=True)
            for mjd, att in attitude.groupby('mjdorbitstate'):
                # Could also use l1.OSTangentPoint([mjd]) to get the tangent point data.
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                osiris_look_eci = ECI(*l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1])
                tangentpoint_eci = ECI(*l1.osl1.LOSTangentPoint(list(position_eci), list(osiris_look_eci), False)[1])
                tangentpoint = Geodetic(*l1.osl1.ECIToGeodetic(mjd, list(tangentpoint_eci))[1])
                self.assertAlmostEqual(tangentpoint.latitude, float(att.osiris_los_latitude), delta=2,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(tangentpoint.longitude, float(att.osiris_los_longitude), delta=2,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(tangentpoint.height, float(att.osiris_los_height), delta=5,
                                       msg=f'ScanNo: {scanno}')

    def test_level1_attitude_quaternions(self):
        """ Test attitude quaternions compared to the quaternions returned by c++ level1 services.
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
            for norm in attitude.Qachieved.pipe(lambda x: x ** 2.0).sum(dim='1ijk').values:
                self.assertAlmostEqual(norm, 1.0, places=6, msg=f'ScanNo: {scanno}')

    def test_level1_housekeeping(self):
        """ Test attitude scan info data compared to the scan info returned by c++ level1 services.
        """
        scanno = 6432012
        hk = open_level1_housekeeping(Orbit(scanno // 1000), scanno=scanno)
        vars = ['stw', 'plus_28v', 'milliamps', 'plus_5v', 'plus_12v', 'plus_24v', 'plus_36v', 'neg_5v', 'neg_12v',
                'gnd_mux1', 'gnd_mux2', 'mjd', 'CPU', 'PS1', 'PS2', 'IRStrap', 'OSStrap', 'IFPlate', 'Optics',
                'IR1', 'IR2', 'IR3', 'CCD', 'RadiatorPlate', 'OS_TecIsOn', 'OS_TecDutyCycle', 'IR_TecIsOn',
                'IR_TecDutyCycle', 'CCD_HtrIsOn', 'IR1_HtrIsOn', 'IR2_HtrIsOn', 'IR3_HtrIsOn', 'CCD_HtrSetPoint',
                'IR1_HtrSetPoint', 'IR2_HtrSetPoint', 'IR3_HtrSetPoint', 'IR_State', 'OS_State', 'IR_SciProg',
                'OS_SciProg', 'SPM_IsBusy', 'TargetIndex', 'LastError', 'NumErrors', 'OSShutterPos', 'IRShutterPos',
                'IRShutterDriverIsOn', 'OSShutterDriverIsOn', 'IRLamp1IsOn', 'IRLamp2IsOn', 'SEUTriggerOccured',
                'SEUDetectorIsActive', 'SEUCounter', 'FifoStatus', 'FifoOverrun', 'FormatsAreLost',
                'MassMemoryIsAvailable', 'MassMemoryCounter', 'SciWordsPerFormat', 'DataIsPendingInSciBuf',
                'UpperFlashOk', 'LowerFlashOk', 'FlashLoadOk', 'InOrbitEnabled', 'VerboseMessagesEnabled',
                'TraceEnabled', 'UsingACDC1', 'UsingACDC2', 'BootSource', 'fswrev', 'ROM_Checksum',
                'LowerFlash_Checksum', 'UpperFlash_Checksum', 'osiris_stw', 'LastCommand', 'NumCommandWordsPending',
                'NumBadCommands', 'NumMarginalCommands', 'NumLostCommandwords', 'NumTotalCommands', 'statstw',
                'statsnumpoints', 'plus5v_minval', 'plus5v_maxval', 'plus5v_average', 'plus12v_minval',
                'plus12v_maxval', 'plus12v_average', 'plus24v_minval', 'plus24v_maxval', 'plus24v_average',
                'plus36v_minval', 'plus36v_maxval', 'plus36v_average', 'neg5v_minval', 'neg5v_maxval',
                'neg5v_average', 'neg12v_minval', 'neg12v_maxval', 'neg12v_average', 'gnd_mux1_minval',
                'gnd_mux1_maxval', 'gnd_mux1_average', 'gnd_mux2_minval', 'gnd_mux2_maxval', 'gnd_mux2_average',
                'statmjd', 'CPU_minval', 'CPU_maxval', 'CPU_average', 'PS1_minval', 'PS1_maxval', 'PS1_average',
                'PS2_minval', 'PS2_maxval', 'PS2_average', 'IRStrap_minval', 'IRStrap_maxval', 'IRStrap_average',
                'OSStrap_minval', 'OSStrap_maxval', 'OSStrap_average', 'IFPlate_minval', 'IFPlate_maxval',
                'IFPlate_average', 'Optics_minval', 'Optics_maxval', 'Optics_average', 'IR1_minval', 'IR1_maxval',
                'IR1_average', 'IR2_minval', 'IR2_maxval', 'IR2_average', 'IR3_minval', 'IR3_maxval', 'IR3_average',
                'CCD_minval', 'CCD_maxval', 'CCD_average', 'dettempstw', 'dettempCCD_T', 'dettempIR1_T',
                'dettempIR2_T', 'dettempIR3_T', 'dettempmjd']
        self.assertSetEqual(set(vars), set(hk.variables), msg=f'ScanNo: {scanno}')

    def test_limit_to_scan_raises_no_spectrograph_data(self):
        """ Test that a KeyError is raised when requesting a scan that has attitude data but no corresponding
            spectrograph data.
        """
        with self.assertRaises(KeyError):
            open_level1_spectrograph(scanno=72176040)

    def test_limit_to_scan_raises_invalid_scan(self):
        """ Test that a KeyError is raised when requesting a scan that doesn't exist.
        """
        with self.assertRaises(KeyError):
            open_level1_spectrograph(scanno=72176099)

    def test_level1_spectrograph_data(self):
        """ Test the spectrograph data compared to the spectrograph data returned by c++ level1 services.
            See the comment in the test code about the c++ not correctly implementing the level1 service.
        """
        for scanno in self.scannos:
            _, spectra, _, _, _ = l1.LoadScanOS(scanno)
            # Replace 0's in the spectra (due to blanked out pixels and radiation hits) with NaN's
            spectra[spectra == 0] = np.NaN
            # OSIRIS Level 1 Services does not blank out severe pixels, only datamissing pixels
            # This is an issue as it is contrary to the Level 1 Services documentation.
            # It behaves properly in matlab by default.
            # Therefore, a different valid spectrograph mask is run on this test than the default.
            spectrograph = open_level1_spectrograph(scanno=scanno, valid=False, load_audit=True)
            valid_spectrograph = spectrograph.where(spectrograph.flags % 2 == 0) \
                .where(spectrograph.radiationhit < MIN_RADIATION_HIT) \
                .where(spectrograph.exceptions == 0)
            np.testing.assert_allclose(valid_spectrograph.data.T, spectra, err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_level1_spectrograph_error(self):
        """ Test attitude spectrograph error compared to the spectrograph error returned by c++ level1 services.
        """
        for scanno in self.scannos:
            _, _, _, _, error = l1.LoadScanOS(scanno)
            spectrograph = open_level1_spectrograph(scanno=scanno)
            np.testing.assert_allclose(spectrograph.error.T, error, err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_spectrograph_wavelengths(self):
        """ Test spectrograph wavelengths compared to the spectrograph wavelengths returned by c++ level1 services.
        """
        for scanno in self.scannos:
            spectrograph = open_level1_spectrograph(scanno=scanno)
            wavelengths = [274.9673, 275.3546, 275.742, 276.1294, 276.5169, 276.9043, 277.2919, 277.6794, 278.067,
                           278.4546, 278.8423, 279.23, 279.6177, 280.0055, 280.3933, 280.7811, 281.169, 281.5569,
                           281.9449, 282.3329, 282.7209, 283.109, 283.497, 283.8852, 284.2733, 284.6615, 285.0498,
                           285.438, 285.8263, 286.2147, 286.6031, 286.9915, 287.3799, 287.7684, 288.1569, 288.5455,
                           288.934, 289.3227, 289.7113, 290.1, 290.4887, 290.8775, 291.2662, 291.6551, 292.0439,
                           292.4328, 292.8217, 293.2107, 293.5997, 293.9887, 294.3778, 294.7668, 295.156, 295.5451,
                           295.9343, 296.3235, 296.7128, 297.1021, 297.4914, 297.8807, 298.2701, 298.6595, 299.049,
                           299.4384, 299.828, 300.2175, 300.6071, 300.9967, 301.3863, 301.776, 302.1657, 302.5554,
                           302.9452, 303.335, 303.7248, 304.1147, 304.5046, 304.8945, 305.2844, 305.6744, 306.0644,
                           306.4545, 306.8446, 307.2347, 307.6248, 308.015, 308.4052, 308.7954, 309.1856, 309.5759,
                           309.9662, 310.3566, 310.747, 311.1374, 311.5278, 311.9183, 312.3088, 312.6993, 313.0898,
                           313.4804, 313.871, 314.2617, 314.6524, 315.0431, 315.4338, 315.8245, 316.2153, 316.6061,
                           316.997, 317.3879, 317.7787, 318.1697, 318.5606, 318.9516, 319.3426, 319.7337, 320.1247,
                           320.5158, 320.907, 321.2981, 321.6893, 322.0805, 322.4717, 322.863, 323.2543, 323.6456,
                           324.0369, 324.4283, 324.8197, 325.2111, 325.6026, 325.994, 326.3855, 326.7771, 327.1686,
                           327.5602, 327.9518, 328.3434, 328.7351, 329.1268, 329.5185, 329.9102, 330.302, 330.6938,
                           331.0856, 331.4774, 331.8693, 332.2612, 332.6531, 333.0451, 333.437, 333.829, 334.221,
                           334.6131, 335.0051, 335.3972, 335.7893, 336.1815, 336.5736, 336.9658, 337.358, 337.7503,
                           338.1425, 338.5348, 338.9271, 339.3195, 339.7118, 340.1042, 340.4966, 340.889, 341.2815,
                           341.674, 342.0665, 342.459, 342.8515, 343.2441, 343.6367, 344.0293, 344.4219, 344.8146,
                           345.2073, 345.6, 345.9927, 346.3854, 346.7782, 347.171, 347.5638, 347.9567, 348.3495,
                           348.7424, 349.1353, 349.5282, 349.9212, 350.3141, 350.7071, 351.1001, 351.4932, 351.8862,
                           352.2793, 352.6724, 353.0655, 353.4586, 353.8518, 354.245, 354.6382, 355.0314, 355.4246,
                           355.8179, 356.2112, 356.6045, 356.9978, 357.3912, 357.7845, 358.1779, 358.5713, 358.9647,
                           359.3582, 359.7516, 360.1451, 360.5386, 360.9321, 361.3257, 361.7192, 362.1128, 362.5064,
                           362.9, 363.2937, 363.6873, 364.081, 364.4747, 364.8684, 365.2621, 365.6559, 366.0496,
                           366.4434, 366.8372, 367.231, 367.6249, 368.0187, 368.4126, 368.8065, 369.2004, 369.5943,
                           369.9883, 370.3822, 370.7762, 371.1702, 371.5642, 371.9582, 372.3523, 372.7463, 373.1404,
                           373.5345, 373.9286, 374.3228, 374.7169, 375.1111, 375.5053, 375.8995, 376.2937, 376.6879,
                           377.0821, 377.4764, 377.8707, 378.265, 378.6593, 379.0536, 379.4479, 379.8423, 380.2367,
                           380.6311, 381.0255, 381.4199, 381.8143, 382.2087, 382.6032, 382.9977, 383.3922, 383.7867,
                           384.1812, 384.5757, 384.9703, 385.3649, 385.7594, 386.154, 386.5486, 386.9432, 387.3379,
                           387.7325, 388.1272, 388.5219, 388.9166, 389.3113, 389.706, 390.1007, 390.4954, 390.8902,
                           391.285, 391.6798, 392.0745, 392.4694, 392.8642, 393.259, 393.6539, 394.0487, 394.4436,
                           394.8385, 395.2334, 395.6283, 396.0232, 396.4181, 396.8131, 397.208, 397.603, 397.998,
                           398.393, 398.788, 399.183, 399.578, 399.973, 400.3681, 400.7632, 401.1582, 401.5533,
                           401.9484, 402.3435, 402.7386, 403.1337, 403.5289, 403.924, 404.3192, 404.7143, 405.1095,
                           405.5047, 405.8999, 406.2951, 406.6903, 407.0856, 407.4808, 407.876, 408.2713, 408.6666,
                           409.0618, 409.4571, 409.8524, 410.2477, 410.643, 411.0384, 411.4337, 411.829, 412.2244,
                           412.6197, 413.0151, 413.4105, 413.8059, 414.2013, 414.5967, 414.9921, 415.3875, 415.7829,
                           416.1783, 416.5738, 416.9692, 417.3647, 417.7601, 418.1556, 418.5511, 418.9466, 419.3421,
                           419.7376, 420.1331, 420.5286, 420.9241, 421.3196, 421.7152, 422.1107, 422.5063, 422.9018,
                           423.2974, 423.693, 424.0885, 424.4841, 424.8797, 425.2753, 425.6709, 426.0665, 426.4621,
                           426.8577, 427.2534, 427.649, 428.0446, 428.4403, 428.8359, 429.2316, 429.6272, 430.0229,
                           430.4185, 430.8142, 431.2099, 431.6056, 432.0012, 432.3969, 432.7926, 433.1883, 433.584,
                           433.9797, 434.3755, 434.7712, 435.1669, 435.5626, 435.9583, 436.3541, 436.7498, 437.1456,
                           437.5413, 437.937, 438.3328, 438.7286, 439.1243, 439.5201, 439.9158, 440.3116, 440.7074,
                           441.1031, 441.4989, 441.8947, 442.2905, 442.6863, 443.082, 443.4778, 443.8736, 444.2694,
                           444.6652, 445.061, 445.4568, 445.8526, 446.2484, 446.6442, 447.04, 447.4358, 447.8316,
                           448.2274, 448.6233, 449.0191, 449.4149, 449.8107, 450.2065, 450.6023, 450.9982, 451.394,
                           451.7898, 452.1856, 452.5814, 452.9773, 453.3731, 453.7689, 454.1647, 454.5605, 454.9564,
                           455.3522, 455.748, 456.1438, 456.5396, 456.9355, 457.3313, 457.7271, 458.1229, 458.5187,
                           458.9146, 459.3104, 459.7062, 460.102, 460.4978, 460.8936, 461.2895, 461.6853, 462.0811,
                           462.4769, 462.8727, 463.2685, 463.6643, 464.0601, 464.4559, 464.8517, 465.2475, 465.6433,
                           466.0391, 466.4348, 466.8306, 467.2264, 467.6222, 468.018, 468.4138, 468.8095, 469.2053,
                           469.6011, 469.9968, 470.3926, 470.7883, 471.1841, 471.5798, 471.9756, 472.3713, 472.7671,
                           473.1628, 473.5585, 473.9543, 474.35, 474.7457, 475.1414, 475.5371, 475.9328, 476.3285,
                           476.7242, 477.1199, 477.5156, 477.9113, 478.307, 478.7027, 479.0983, 479.494, 479.8897,
                           480.2853, 480.681, 481.0766, 481.4723, 481.8679, 482.2635, 482.6591, 483.0548, 483.4504,
                           483.846, 484.2416, 484.6372, 485.0327, 485.4283, 485.8239, 486.2195, 486.615, 487.0106,
                           487.4061, 487.8017, 488.1972, 488.5927, 488.9883, 489.3838, 489.7793, 490.1748, 490.5703,
                           490.9658, 491.3612, 491.7567, 492.1522, 492.5476, 492.9431, 493.3385, 493.7339, 494.1294,
                           494.5248, 494.9202, 495.3156, 495.711, 496.1063, 496.5017, 496.8971, 497.2924, 497.6878,
                           498.0831, 498.4785, 498.8738, 499.2691, 499.6644, 500.0597, 500.455, 500.8502, 501.2455,
                           501.6408, 502.036, 502.4312, 502.8265, 503.2217, 503.6169, 504.0121, 504.4073, 504.8024,
                           505.1976, 505.5928, 505.9879, 506.383, 506.7782, 507.1733, 507.5684, 507.9635, 508.3585,
                           508.7536, 509.1487, 509.5437, 509.9388, 510.3338, 510.7288, 511.1238, 511.5188, 511.9138,
                           512.3087, 512.7037, 513.0986, 513.4936, 513.8885, 514.2834, 514.6783, 515.0732, 515.468,
                           515.8629, 516.2577, 516.6526, 517.0474, 517.4422, 517.837, 518.2317, 518.6265, 519.0213,
                           519.416, 519.8107, 520.2054, 520.6001, 520.9948, 521.3895, 521.7842, 522.1788, 522.5734,
                           522.9681, 523.3627, 523.7573, 524.1518, 524.5464, 524.9409, 525.3355, 525.73, 526.1245,
                           526.519, 526.9135, 527.3079, 527.7024, 528.0968, 528.4912, 528.8856, 529.28, 529.6743,
                           530.0687, 530.463, 531.829, 532.2293, 532.6295, 533.0298, 533.43, 533.8303, 534.2306,
                           534.6308, 535.0311, 535.4314, 535.8317, 536.232, 536.6324, 537.0327, 537.433, 537.8334,
                           538.2337, 538.6341, 539.0345, 539.4348, 539.8352, 540.2356, 540.636, 541.0364, 541.4369,
                           541.8373, 542.2377, 542.6382, 543.0386, 543.4391, 543.8395, 544.24, 544.6405, 545.041,
                           545.4415, 545.842, 546.2425, 546.643, 547.0436, 547.4441, 547.8446, 548.2452, 548.6458,
                           549.0463, 549.4469, 549.8475, 550.2481, 550.6487, 551.0493, 551.4499, 551.8506, 552.2512,
                           552.6518, 553.0525, 553.4532, 553.8538, 554.2545, 554.6552, 555.0559, 555.4566, 555.8573,
                           556.258, 556.6587, 557.0595, 557.4602, 557.8609, 558.2617, 558.6625, 559.0632, 559.464,
                           559.8648, 560.2656, 560.6664, 561.0672, 561.4681, 561.8689, 562.2697, 562.6706, 563.0714,
                           563.4723, 563.8732, 564.274, 564.6749, 565.0758, 565.4767, 565.8776, 566.2785, 566.6795,
                           567.0804, 567.4814, 567.8823, 568.2833, 568.6842, 569.0852, 569.4862, 569.8872, 570.2882,
                           570.6892, 571.0902, 571.4912, 571.8923, 572.2933, 572.6943, 573.0954, 573.4965, 573.8975,
                           574.2986, 574.6997, 575.1008, 575.5019, 575.903, 576.3041, 576.7053, 577.1064, 577.5076,
                           577.9087, 578.3099, 578.711, 579.1122, 579.5134, 579.9146, 580.3158, 580.717, 581.1182,
                           581.5195, 581.9207, 582.3219, 582.7232, 583.1244, 583.5257, 583.927, 584.3283, 584.7295,
                           585.1308, 585.5322, 585.9335, 586.3348, 586.7361, 587.1375, 587.5388, 587.9402, 588.3415,
                           588.7429, 589.1443, 589.5457, 589.9471, 590.3485, 590.7499, 591.1513, 591.5527, 591.9542,
                           592.3556, 592.7571, 593.1585, 593.56, 593.9615, 594.3629, 594.7644, 595.1659, 595.5674,
                           595.969, 596.3705, 596.772, 597.1736, 597.5751, 597.9767, 598.3782, 598.7798, 599.1814,
                           599.583, 599.9846, 600.3862, 600.7878, 601.1894, 601.591, 601.9927, 602.3943, 602.796,
                           603.1976, 603.5993, 604.001, 604.4027, 604.8044, 605.2061, 605.6078, 606.0095, 606.4112,
                           606.813, 607.2147, 607.6165, 608.0182, 608.42, 608.8218, 609.2235, 609.6253, 610.0271,
                           610.4289, 610.8308, 611.2326, 611.6344, 612.0362, 612.4381, 612.84, 613.2418, 613.6437,
                           614.0456, 614.4475, 614.8494, 615.2513, 615.6532, 616.0551, 616.457, 616.859, 617.2609,
                           617.6629, 618.0648, 618.4668, 618.8688, 619.2707, 619.6727, 620.0747, 620.4767, 620.8788,
                           621.2808, 621.6828, 622.0849, 622.4869, 622.889, 623.291, 623.6931, 624.0952, 624.4973,
                           624.8994, 625.3015, 625.7036, 626.1057, 626.5079, 626.91, 627.3121, 627.7143, 628.1165,
                           628.5186, 628.9208, 629.323, 629.7252, 630.1274, 630.5296, 630.9318, 631.3341, 631.7363,
                           632.1385, 632.5408, 632.943, 633.3453, 633.7476, 634.1499, 634.5522, 634.9545, 635.3568,
                           635.7591, 636.1614, 636.5637, 636.9661, 637.3684, 637.7708, 638.1732, 638.5755, 638.9779,
                           639.3803, 639.7827, 640.1851, 640.5875, 640.9899, 641.3924, 641.7948, 642.1973, 642.5997,
                           643.0022, 643.4046, 643.8071, 644.2096, 644.6121, 645.0146, 645.4171, 645.8196, 646.2222,
                           646.6247, 647.0272, 647.4298, 647.8323, 648.2349, 648.6375, 649.0401, 649.4427, 649.8453,
                           650.2479, 650.6505, 651.0531, 651.4557, 651.8584, 652.261, 652.6637, 653.0663, 653.469,
                           653.8717, 654.2744, 654.6771, 655.0798, 655.4825, 655.8852, 656.2879, 656.6907, 657.0934,
                           657.4962, 657.8989, 658.3017, 658.7045, 659.1073, 659.5101, 659.9129, 660.3157, 660.7185,
                           661.1213, 661.5241, 661.927, 662.3298, 662.7327, 663.1356, 663.5384, 663.9413, 664.3442,
                           664.7471, 665.15, 665.5529, 665.9558, 666.3588, 666.7617, 667.1647, 667.5676, 667.9706,
                           668.3735, 668.7765, 669.1795, 669.5825, 669.9855, 670.3885, 670.7915, 671.1946, 671.5976,
                           672.0006, 672.4037, 672.8068, 673.2098, 673.6129, 674.016, 674.4191, 674.8222, 675.2253,
                           675.6284, 676.0315, 676.4346, 676.8378, 677.2409, 677.6441, 678.0473, 678.4504, 678.8536,
                           679.2568, 679.66, 680.0632, 680.4664, 680.8696, 681.2729, 681.6761, 682.0793, 682.4826,
                           682.8859, 683.2891, 683.6924, 684.0957, 684.499, 684.9023, 685.3056, 685.7089, 686.1122,
                           686.5156, 686.9189, 687.3223, 687.7256, 688.129, 688.5324, 688.9358, 689.3391, 689.7425,
                           690.146, 690.5494, 690.9528, 691.3562, 691.7597, 692.1631, 692.5666, 692.97, 693.3735,
                           693.777, 694.1805, 694.584, 694.9875, 695.391, 695.7945, 696.198, 696.6016, 697.0051,
                           697.4087, 697.8122, 698.2158, 698.6194, 699.0229, 699.4265, 699.8301, 700.2337, 700.6374,
                           701.041, 701.4446, 701.8483, 702.2519, 702.6556, 703.0592, 703.4629, 703.8666, 704.2703,
                           704.674, 705.0777, 705.4814, 705.8851, 706.2888, 706.6926, 707.0963, 707.5001, 707.9038,
                           708.3076, 708.7114, 709.1152, 709.519, 709.9228, 710.3266, 710.7304, 711.1342, 711.5381,
                           711.9419, 712.3458, 712.7496, 713.1535, 713.5574, 713.9613, 714.3651, 714.769, 715.1729,
                           715.5769, 715.9808, 716.3847, 716.7887, 717.1926, 717.5966, 718.0005, 718.4045, 718.8085,
                           719.2125, 719.6165, 720.0205, 720.4245, 720.8285, 721.2325, 721.6366, 722.0406, 722.4447,
                           722.8487, 723.2528, 723.6569, 724.061, 724.4651, 724.8692, 725.2733, 725.6774, 726.0815,
                           726.4856, 726.8898, 727.2939, 727.6981, 728.1023, 728.5064, 728.9106, 729.3148, 729.719,
                           730.1232, 730.5274, 730.9316, 731.3359, 731.7401, 732.1444, 732.5486, 732.9529, 733.3572,
                           733.7614, 734.1657, 734.57, 734.9743, 735.3786, 735.783, 736.1873, 736.5916, 736.996,
                           737.4003, 737.8047, 738.209, 738.6134, 739.0178, 739.4222, 739.8266, 740.231, 740.6354,
                           741.0398, 741.4443, 741.8487, 742.2532, 742.6576, 743.0621, 743.4666, 743.871, 744.2755,
                           744.68, 745.0845, 745.489, 745.8936, 746.2981, 746.7026, 747.1072, 747.5117, 747.9163,
                           748.3209, 748.7254, 749.13, 749.5346, 749.9392, 750.3438, 750.7485, 751.1531, 751.5577,
                           751.9624, 752.367, 752.7717, 753.1763, 753.581, 753.9857, 754.3904, 754.7951, 755.1998,
                           755.6045, 756.0092, 756.414, 756.8187, 757.2235, 757.6282, 758.033, 758.4377, 758.8425,
                           759.2473, 759.6521, 760.0569, 760.4617, 760.8665, 761.2714, 761.6762, 762.0811, 762.4859,
                           762.8908, 763.2956, 763.7005, 764.1054, 764.5103, 764.9152, 765.3201, 765.725, 766.1299,
                           766.5349, 766.9398, 767.3448, 767.7497, 768.1547, 768.5597, 768.9646, 769.3696, 769.7746,
                           770.1796, 770.5847, 770.9897, 771.3947, 771.7997, 772.2048, 772.6099, 773.0149, 773.42,
                           773.8251, 774.2302, 774.6352, 775.0403, 775.4455, 775.8506, 776.2557, 776.6608, 777.066,
                           777.4711, 777.8763, 778.2815, 778.6866, 779.0918, 779.497, 779.9022, 780.3074, 780.7126,
                           781.1179, 781.5231, 781.9283, 782.3336, 782.7388, 783.1441, 783.5494, 783.9547, 784.36,
                           784.7652, 785.1706, 785.5759, 785.9812, 786.3865, 786.7919, 787.1972, 787.6026, 788.0079,
                           788.4133, 788.8187, 789.224, 789.6294, 790.0348, 790.4403, 790.8457, 791.2511, 791.6565,
                           792.062, 792.4674, 792.8729, 793.2784, 793.6838, 794.0893, 794.4948, 794.9003, 795.3058,
                           795.7113, 796.1168, 796.5224, 796.9279, 797.3335, 797.739, 798.1446, 798.5502, 798.9557,
                           799.3613, 799.7669, 800.1725, 800.5781, 800.9838, 801.3894, 801.795, 802.2007, 802.6063,
                           803.012, 803.4176, 803.8233, 804.229, 804.6347, 805.0404, 805.4461, 805.8518, 806.2576,
                           806.6633, 807.069, 807.4748, 807.8805, 808.2863, 808.6921, 809.0979, 809.5036, 809.9094,
                           810.3152, 810.7211, 811.1269, 811.5327, 811.9386, 812.3444, 812.7503, 813.1561, 813.562,
                           813.9679, 814.3738, 814.7796]
            np.testing.assert_allclose(spectrograph.wavelength, wavelengths, atol=4e-5, err_msg=f'ScanNo: {scanno}',
                                       verbose=True)

    def test_spectrograph_fwhm(self):
        """ Test spectrograph fwhm values compared to the values in the CDB level1 services directory.
        """
        psf = HDF4VS('/media/odp/OsirisConfigurationDatabase/cdb/ospxcopy.hdf')['R8:R4:01353']['Array'][3]
        for scanno in self.scannos:
            spectrograph = open_level1_spectrograph(scanno=scanno)
            np.testing.assert_allclose(spectrograph.fwhm, psf, atol=1e-6, err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_level1_ir(self):
        """ Test that ir level1 data reads and returns the proper variables.
        """
        for scanno in [6432012]:
            ir = open_level1_ir(channel=1, scanno=scanno, load_audit=True)
            data_vars = ['stw', 'exposureTime', 'temperature', 'tempavg', 'mode', 'scienceprog', 'shutter', 'lamp1',
                         'lamp2', 'targetIndex', 'exceptions', 'processingflags', 'data', 'error', 'flags', 'audit']
            self.assertSetEqual(set(data_vars), set(ir.data_vars), msg=f'ScanNo: {scanno}')

    def test_level1b_ir(self):
        """ Test that ir level1b data reads and returns the proper variables.
        """
        for scanno in [6432012]:
            ir = open_level1b_ir(channel=1, scanno=scanno, load_audit=True)
            data_vars = ['stw', 'exposureTime', 'temperature', 'tempavg', 'detectorId', 'mode', 'scienceprog',
                         'shutter', 'lamp1', 'lamp2', 'targetIndex', 'exceptions', 'processingflags', 'data', 'error',
                         'flags', 'audit']
            self.assertSetEqual(set(data_vars), set(ir.data_vars), msg=f'ScanNo: {scanno}')

    def test_level1_ecmwf(self):
        """ Test the OSIRIS ecmwf ERA-Interim compared to the ERA-Interim data in the L2V5 database.
        """
        scanno = 6432012
        # The temperature and pressure profiles for scan 6432012 taken from the L2V5 database on the half km grid
        density = [2.6524082e+19, 2.3140975e+19, 2.046554e+19, 1.8377619e+19, 1.6451248e+19, 1.467451e+19,
                   1.3040925e+19, 1.1538201e+19, 1.0105009e+19, 8.7356072e+18, 7.5134187e+18, 6.4258544e+18,
                   5.5115087e+18, 4.740722e+18, 4.07735608e+18, 3.49842885e+18, 2.99664802e+18, 2.56827637e+18,
                   2.20402438e+18, 1.89070604e+18, 1.62094911e+18, 1.38858024e+18, 1.18799744e+18, 1.01144054e+18,
                   8.5902266e+17, 7.2898597e+17, 6.1868743e+17, 5.26312405e+17, 4.48104899e+17, 3.81715184e+17,
                   3.25001447e+17, 2.77739609e+17, 2.3805394e+17, 2.04848929e+17, 1.76433821e+17, 1.51874201e+17,
                   1.30476245e+17, 1.12325464e+17, 9.6797405e+16, 8.3568244e+16, 7.2087942e+16, 6.2094713e+16,
                   5.3622117e+16, 4.6387133e+16, 4.0327059e+16, 3.52233854e+16, 3.07724679e+16, 2.70237367e+16,
                   2.37383293e+16, 2.08539367e+16, 1.84963132e+16, 1.64109358e+16, 1.45612319e+16, 1.29293956e+16,
                   1.150164e+16, 1.02325126e+16, 9104292400000000.0, 8101245200000000.0, 7171793300000000.0,
                   6337694000000000.0, 5601418000000000.0, 4951410100000000.0, 4377487580000000.0, 3870678210000000.0,
                   3423074710000000.0, 3027707400000000.0, 2717751420000000.0, 2423680110000000.0, 2158448570000000.0,
                   1919462360000000.0, 1704225310000000.0, 1510917150000000.0, 1326594740000000.0, 1155320700000000.0,
                   1006615780000000.0, 875437080000000.0, 759737100000000.0, 657437220000000.0, 567021380000000.0,
                   487138815000000.0, 416995120000000.0, 355421160000000.0, 301374667000000.0, 254400275000000.0,
                   213736380000000.0, 178750113000000.0, 148823251000000.0, 123379303000000.0, 101876297000000.0,
                   83810448000000.0, 68718407000000.0, 56184338000000.0, 45834423000000.0, 37329461000000.0,
                   30368732300000.0, 24690473000000.0, 20070342300000.0, 16318500700000.0, 13275976700000.0,
                   10810754600000.0]
        temperature = [260.30127, 261.66544, 259.46716, 252.78905, 246.19312, 239.76208, 233.51962, 227.55983,
                       223.29218, 221.52238, 220.72989, 221.17393, 221.02106, 220.16837, 219.20557, 218.67604,
                       218.47186, 218.10657, 217.38464, 216.65358, 215.93092, 215.28743, 214.85448, 215.49481,
                       216.81555, 218.53545, 220.52455, 222.43619, 224.40877, 226.55696, 229.38602, 231.74571,
                       233.63748, 234.78827, 236.16264, 237.89023, 240.31174, 242.79517, 245.32732, 247.96135, 250.8831,
                       254.22812, 258.14844, 262.43057, 266.24045, 269.5616, 272.86536, 274.25177, 275.57843, 276.90265,
                       276.79126, 276.63782, 276.4844, 275.78384, 273.7054, 271.62766, 269.5505, 267.47403, 264.93588,
                       262.2322, 259.52936, 256.82736, 254.12622, 251.4259, 248.72644, 246.0278, 242.03139, 238.36552,
                       234.61807, 230.79004, 226.88417, 222.90111, 219.19788, 215.59056, 211.6247, 207.42824, 203.10326,
                       198.74413, 194.43053, 190.23001, 186.19862, 182.38263, 178.82004, 175.54231, 172.57593,
                       169.94398, 167.66766, 165.76762, 164.26541, 163.18495, 162.55312, 162.38113, 162.66612, 163.4121,
                       164.6309, 166.34288, 168.57759, 171.37524, 174.78864, 178.88565]
        ecmwf = open_level1_ecmwf(scanno=scanno)
        np.testing.assert_allclose(ecmwf.sel(Altitude=np.linspace(500.0, 99500.0, 100)).temperature, temperature,
                                   err_msg=f'ScanNo: {scanno}', verbose=True)
        np.testing.assert_allclose(ecmwf.sel(Altitude=np.linspace(500.0, 99500.0, 100)).density.values, density,
                                   err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_level1_ecmwf_no_data(self):
        """ Test that orbits with a mix of scans with/without OSIRIS ecmwf ERA-Interim data load properly.
        """
        for scanno in [6424005, 7049005, 7054005, 7059005, 7104005, 7109005, 7014005, 7179005, 7180005, 89514005]:
            open_level1_ecmwf(scanno=scanno)

    def test_level1_ecmwf_empty_file(self):
        """ Test that orbits with an empty OSIRIS ecmwf ERA-Interim file throw an exception.
        """
        with self.assertRaises(KeyError):
            open_level1_ecmwf(scanno=7019005)

    def test_level1_files_exist(self):
        """ Test that scans have the necessary level1 data to load properly, but do not actually read the data in.
        """
        from collections import Counter

        orbitlist = np.loadtxt(os.path.join(os.path.dirname(__file__), 'data', 'orbitsforfilecheck.txt'), dtype='int')

        file_error_count = Counter()
        error_dict: DefaultDict[str, List[int]] = defaultdict(list)

        for orbit in orbitlist:
            for filetype in Orbit(orbit).l1dict:
                if filetype.startswith('IR'):  # Ignore IR files as they are not currently processed
                    continue
                try:
                    Orbit(orbit)._l1_filename(filetype)
                except LookupError as e:  # pragma: no cover
                    print(e)
                    error_dict[filetype].append(orbit)
                    file_error_count[filetype] += 1

        if file_error_count:  # pragma: no cover
            print(error_dict)
            self.fail(file_error_count)
